﻿using Microsoft.EntityFrameworkCore;
using PiwigoDirectorySync.Infrastructure;
using PiwigoDirectorySync.Persistence;
using Serilog;
using Spectre.Console;
using Spectre.Console.Cli;

namespace PiwigoDirectorySync.Commands;

internal class PiwigoAddCommand : CancellableAsyncCommand<PiwigoAddCommand.PiwigoAddSettings>
{
    private readonly ILogger _logger = Log.ForContext<PiwigoAddCommand>();
    private readonly PersistenceContext _persistenceContext;

    public PiwigoAddCommand(PersistenceContext persistenceContext)
    {
        _persistenceContext = persistenceContext;
    }

    protected override async Task<int> ExecuteAsync(CommandContext context, PiwigoAddSettings settings, CancellationToken cancellation)
    {
        _logger.Information("Adding piwigo server {SettingsServerName} with user {SettingsUserName} and url {SettingsUrl}", settings.ServerName, settings.UserName, settings.Url);

        var count = await _persistenceContext.PiwigoServers.Where(s => s.Name == settings.ServerName).CountAsync(cancellation);
        if (count > 0)
        {
            _logger.Warning("Server with name {SettingsServerName} already exists. Not updated", settings.ServerName);
            AnsiConsole.MarkupLine($"[maroon]There is already a server with the name {settings.ServerName}[/]");
            return -1;
        }

        var server = new ServerEntity
        {
            Url = settings.Url,
            Name = settings.ServerName,
            Username = settings.UserName,
            Password = settings.Password,
            RootDirectory = settings.RootDirectory
        };

        _persistenceContext.PiwigoServers.Add(server);
        await _persistenceContext.SaveChangesAsync(cancellation);

        AnsiConsole.MarkupLine($"[green]Added server {settings.ServerName} with id {server.Id}[/]");

        var piwigoServers = await _persistenceContext.PiwigoServers.ToListAsync(cancellation);
        PiwigoListCommand.PrintServers(piwigoServers);

        return 0;
    }

    internal class PiwigoAddSettings : CommandSettings
    {
        [CommandArgument(0, "<ServerName>")] public string ServerName { get; init; } = string.Empty;
        [CommandArgument(1, "<UserName>")] public string UserName { get; init; } = string.Empty;
        [CommandArgument(2, "<Password>")] public string Password { get; init; } = string.Empty;
        [CommandArgument(3, "<Url>")] public string Url { get; init; } = string.Empty;
        [CommandArgument(4, "<RootDirectory>")] public string RootDirectory { get; init; } = string.Empty;

        public override ValidationResult Validate()
        {
            var baseResult = base.Validate();
            if (!baseResult.Successful)
            {
                return baseResult;
            }

            if (string.IsNullOrWhiteSpace(ServerName))
            {
                return ValidationResult.Error("Please specify a server name");
            }

            if (string.IsNullOrWhiteSpace(Url))
            {
                return ValidationResult.Error("Please specify a server url");
            }

            if (string.IsNullOrWhiteSpace(UserName))
            {
                return ValidationResult.Error("Please specify a valid username for the server");
            }

            if (string.IsNullOrWhiteSpace(Password))
            {
                return ValidationResult.Error("Please specify a password");
            }

            if (string.IsNullOrWhiteSpace(RootDirectory))
            {
                return ValidationResult.Error("Please specify a root directory");
            }

            if (!Directory.Exists(RootDirectory))
            {
                return ValidationResult.Error($"The root directory {RootDirectory} does not exist or is not accessible");
            }

            return ValidationResult.Success();
        }
    }
}