﻿using Microsoft.EntityFrameworkCore;
using PiwigoDirectorySync.Infrastructure;
using PiwigoDirectorySync.Persistence;
using Serilog;
using Spectre.Console;
using Spectre.Console.Cli;

namespace PiwigoDirectorySync.Commands;

internal class PiwigoRemoveCommand : CancellableAsyncCommand<PiwigoRemoveCommand.PiwigoRemoveSettings>
{
    private readonly ILogger _logger = Log.ForContext<PiwigoRemoveCommand>();
    private readonly PersistenceContext _persistenceContext;

    public PiwigoRemoveCommand(PersistenceContext persistenceContext)
    {
        _persistenceContext = persistenceContext;
    }

    protected override async Task<int> ExecuteAsync(CommandContext context, PiwigoRemoveSettings settings, CancellationToken cancellation)
    {
        var server = await _persistenceContext.PiwigoServers.FindByIdAsync(settings.PiwigoServerId, cancellation);
        if (server is null)
        {
            _logger.Warning("Could not find server with id {SettingsPiwigoServerId}", settings.PiwigoServerId);
            AnsiConsole.MarkupLine($"[maroon]Could not find server with id {settings.PiwigoServerId}[/]");
            return -1;
        }

        var serverName = server.Name;
        var serverId = server.Id;
        if (!AnsiConsole.Confirm($"[orange3]Do you really want to remove {serverName} ({serverId}) from this database?[/]", false))
        {
            _logger.Information("Do not remove server {ServerName}", serverName);
            AnsiConsole.MarkupLine("[green]Remove cancelled![/]");
            return 1;
        }

        _logger.Information("Removing {ServerName} including images and albums from database", serverName);
        
        await AnsiConsole.Status()
            .Spinner(Spinner.Known.Star)
            .SpinnerStyle(Style.Parse("green bold"))
            .StartAsync("[orange3]Updating database[/]", async ctx =>
            {
                ctx.Status($"[orange3]Removing image metadata of server {serverName}![/]");
                var deletedImages = await _persistenceContext.PiwigoImages.Where(i => i.Album.ServerId == serverId).ExecuteDeleteAsync(cancellation);
                await _persistenceContext.SaveChangesAsync(cancellation);
                ctx.Status($"[green]Deleted {deletedImages} images from database[/]");

                ctx.Status($"[orange3]Removing album metadata of server {serverName}![/]");
                var deletedAlbums = await _persistenceContext.PiwigoAlbums.Where(a => a.ServerId == serverId).ExecuteDeleteAsync(cancellation);
                await _persistenceContext.SaveChangesAsync(cancellation);
                ctx.Status($"[green]Deleted {deletedAlbums} albums from database[/]");

                ctx.Status($"[orange3]Removing server {serverName}![/]");
                await _persistenceContext.PiwigoServers.Where(s => s.Id == serverId).ExecuteDeleteAsync(cancellation);
                await _persistenceContext.SaveChangesAsync(cancellation);
                ctx.Status($"[green]Deleted piwigo server {serverName}[/]");
            });

        _logger.Information("Removed {ServerName} including images and albums from database", serverName);

        return 0;
    }

    internal class PiwigoRemoveSettings : CommonCommandSettings
    {
    }
}