﻿using System.Diagnostics.CodeAnalysis;
using Microsoft.EntityFrameworkCore;
using PiwigoDirectorySync.Infrastructure;
using PiwigoDirectorySync.Persistence;
using Serilog;
using Spectre.Console;
using Spectre.Console.Cli;

namespace PiwigoDirectorySync.Commands;

internal class PiwigoUpdateCommand : CancellableAsyncCommand<PiwigoUpdateCommand.PiwigoUpdateSettings>
{
    private readonly ILogger _logger = Log.ForContext<PiwigoUpdateCommand>();

    private readonly PersistenceContext _persistenceContext;

    public PiwigoUpdateCommand(PersistenceContext persistenceContext)
    {
        _persistenceContext = persistenceContext;
    }

    protected override async Task<int> ExecuteAsync(CommandContext context, PiwigoUpdateSettings settings, CancellationToken cancellation)
    {
        _logger.Information("Updating piwigo server {SettingsServerName} with user {SettingsUserName} and url {SettingsUrl}", settings.ServerName, settings.UserName, settings.Url);

        var count = await _persistenceContext.PiwigoServers.Where(s => s.Name == settings.ServerName && s.Id != settings.PiwigoServerId).CountAsync(cancellation);
        if (count > 0)
        {
            _logger.Warning("Server with name {SettingsServerName} already exists. Not updated", settings.ServerName);
            AnsiConsole.MarkupLine(
                $"[maroon]You can not rename the server with id {settings.PiwigoServerId} to {settings.ServerName} as there is already a server with this name[/]");
            return -1;
        }

        var server = await _persistenceContext.PiwigoServers.FindByIdAsync(settings.PiwigoServerId, cancellation);
        if (server is null)
        {
            _logger.Warning("Could not find server with id {SettingsPiwigoServerId}", settings.PiwigoServerId);
            AnsiConsole.MarkupLine($"[maroon]Could not find server with id {settings.PiwigoServerId}[/]");
            return -2;
        }

        server.Name = settings.ServerName ?? server.Name;
        server.Username = settings.ServerName ?? server.Username;
        server.Password = settings.Password ?? server.Password;
        server.Url = settings.Url ?? server.Url;
        server.RootDirectory = settings.RootDirectory ?? server.RootDirectory;

        await _persistenceContext.SaveChangesAsync(cancellation);

        AnsiConsole.MarkupLine($"[green]Added server {settings.ServerName} with id {server.Id}[/]");

        var piwigoServers = await _persistenceContext.PiwigoServers.ToListAsync(cancellation);
        PiwigoListCommand.PrintServers(piwigoServers);

        return 0;
    }

    [SuppressMessage("ReSharper", "UnusedAutoPropertyAccessor.Global")]
    internal class PiwigoUpdateSettings : CommonCommandSettings
    {
        [CommandArgument(0, "[ServerName]")] public string? ServerName { get; init; }
        [CommandArgument(1, "[UserName]")] public string? UserName { get; init; }
        [CommandArgument(2, "[Password]")] public string? Password { get; init; }
        [CommandArgument(3, "[Url]")] public string? Url { get; init; }
        [CommandArgument(4, "[RootDirectory]")] public string? RootDirectory { get; init; }

        public override ValidationResult Validate()
        {
            var baseResult = base.Validate();
            if (!baseResult.Successful)
            {
                return baseResult;
            }

            if (string.IsNullOrWhiteSpace(ServerName))
            {
                return ValidationResult.Error("Please specify a server name");
            }

            if (string.IsNullOrWhiteSpace(Url))
            {
                return ValidationResult.Error("Please specify a server url");
            }

            if (string.IsNullOrWhiteSpace(UserName))
            {
                return ValidationResult.Error("Please specify a valid username for the server");
            }

            if (string.IsNullOrWhiteSpace(Password))
            {
                return ValidationResult.Error("Please specify a password");
            }

            if (string.IsNullOrWhiteSpace(RootDirectory))
            {
                return ValidationResult.Error("Please specify a root directory");
            }

            if (!Directory.Exists(RootDirectory))
            {
                return ValidationResult.Error($"The root directory {RootDirectory} does not exist or is not accessible");
            }

            return ValidationResult.Success();
        }
    }
}