﻿using System.ComponentModel;
using System.Diagnostics;
using System.Diagnostics.CodeAnalysis;
using System.Threading.Channels;
using PiwigoDirectorySync.Infrastructure;
using PiwigoDirectorySync.Services;
using Serilog;
using Spectre.Console;
using Spectre.Console.Cli;

namespace PiwigoDirectorySync.Commands;

internal class ScanCommand : CancellableAsyncCommand<ScanCommand.ScanSettings>
{
    private readonly IFileIndexer _fileIndexer;
    private readonly IFileSystemScanner _fileSystemScanner;
    private readonly ILogger _logger = Log.ForContext<ScanCommand>();

    public ScanCommand(IFileIndexer fileIndexer, IFileSystemScanner fileSystemScanner)
    {
        _fileIndexer = fileIndexer;
        _fileSystemScanner = fileSystemScanner;
    }

    protected override async Task<int> ExecuteAsync(CommandContext context, ScanSettings settings, CancellationToken cancellationToken)
    {
        await ScanDirectory(_logger, _fileIndexer, _fileSystemScanner, settings.PiwigoServerId, cancellationToken);
        return 0;
    }

    internal static async Task ScanDirectory(ILogger logger, IFileIndexer fileIndexer, IFileSystemScanner fileSystemScanner, int piwigoServerId, CancellationToken ct)
    {
        //TODO: check files for deletion -> files in db but no longer exist
        logger.Information("Starting scanner and remover");
        var stopWatch = Stopwatch.StartNew();
        var fileQueue = Channel.CreateUnbounded<string>();

        var indexerTask = fileIndexer.StartProcessingAsync(fileQueue, piwigoServerId, ct);

        var scannerTask = fileSystemScanner.ScanAsync(fileQueue, piwigoServerId, ct);
        await Task.WhenAll(scannerTask, indexerTask);

        stopWatch.Stop();
        logger.Information("Processed {IndexerTotalFilesScanned} image files in {ElapsedTotalSeconds} seconds", fileIndexer.TotalFilesScanned, stopWatch.Elapsed.TotalSeconds);

        foreach (var failedFilePath in fileIndexer.FailedFiles)
        {
            AnsiConsole.MarkupLine($"[red]Failed to index file {failedFilePath}[/]");
            logger.Error("Failed to index file {FailedFilePath}", failedFilePath);
        }
    }

    [SuppressMessage("ReSharper", "UnusedAutoPropertyAccessor.Global")]
    internal class ScanSettings : CommonCommandSettings
    {
        [CommandOption("-d|--mark-for-delete")]
        [DefaultValue(false)]
        public bool MarkForDelete { get; set; }
    }
}