using System.Diagnostics;
using PiwigoDirectorySync.Infrastructure;
using PiwigoDirectorySync.Services;
using Serilog;
using Spectre.Console.Cli;

namespace PiwigoDirectorySync.Commands;

internal class SyncFullCommand : CancellableAsyncCommand<SyncFullCommand.SyncFullSettings>
{
    private readonly IAlbumSynchronizer _albumSynchronizer;
    private readonly IFileIndexer _fileIndexer;
    private readonly IFileSystemScanner _fileSystemScanner;
    private readonly IImageSynchronizer _imageSynchronizer;
    private readonly ILogger _logger = Log.ForContext<SyncFullCommand>();

    public SyncFullCommand(IAlbumSynchronizer albumSynchronizer, IFileIndexer fileIndexer, IFileSystemScanner fileSystemScanner, IImageSynchronizer imageSynchronizer)
    {
        _albumSynchronizer = albumSynchronizer;
        _fileIndexer = fileIndexer;
        _fileSystemScanner = fileSystemScanner;
        _imageSynchronizer = imageSynchronizer;
    }

    protected override async Task<int> ExecuteAsync(CommandContext context, SyncFullSettings settings, CancellationToken cancellationToken)
    {
        _logger.Information("Starting full synchronization for piwigo server {SettingsPiwigoServerId}", settings.PiwigoServerId);
        var stopWatch = Stopwatch.StartNew();

        _logger.Information("running file system scan");
        await ScanCommand.ScanDirectory(_logger, _fileIndexer, _fileSystemScanner, settings.PiwigoServerId, cancellationToken);

        _logger.Information("running album synchronization");
        await _albumSynchronizer.SynchronizeAlbums(settings.PiwigoServerId, settings.ReadOnly, cancellationToken);

        _logger.Information("running image synchronization");
        await _imageSynchronizer.SynchronizeImagesAsync(settings.PiwigoServerId, settings.ReadOnly, cancellationToken);

        stopWatch.Stop();
        _logger.Information("Full synchronization for piwigo server {SettingsPiwigoServerId} finished in {ElapsedTotalSeconds} seconds", settings.PiwigoServerId,
            stopWatch.Elapsed.TotalSeconds);

        return 0;
    }

    internal class SyncFullSettings : SyncCommandSettings
    {
    }
}