﻿using Microsoft.EntityFrameworkCore;
using Piwigo.Client;
using Piwigo.Client.Albums;
using PiwigoDirectorySync.Infrastructure;
using PiwigoDirectorySync.Persistence;
using Serilog;

namespace PiwigoDirectorySync.Services;

internal class AlbumSynchronizer : IAlbumSynchronizer
{
    private readonly ILogger _logger = Log.ForContext<AlbumSynchronizer>();
    private readonly PersistenceContext _persistenceContext;
    private readonly IPiwigoClientFactory _piwigoClientFactory;

    public AlbumSynchronizer(IPiwigoClientFactory piwigoClientFactory, PersistenceContext persistenceContext)
    {
        _piwigoClientFactory = piwigoClientFactory;
        _persistenceContext = persistenceContext;
    }

    public async Task SynchronizeAlbums(int piwigoServerId, bool readOnly, CancellationToken ct)
    {
        var piwigoServer = await _persistenceContext.PiwigoServers.FindAsync(new object?[] { piwigoServerId }, ct);
        if (piwigoServer is null)
        {
            _logger.Error("Could not sync albums with piwigo server {PiwigoServerId}", piwigoServerId);
            return;
        }

        var piwigoClient = await _piwigoClientFactory.GetPiwigoClientAsync(piwigoServer, ct);

        await UpdatePiwigoAlbumsFromServerAsync(piwigoClient, piwigoServer, ct);
        if (!readOnly)
        {
            _logger.Information("Creating missing albums on server {PiwigoServerId} - {PiwigoServerName}", piwigoServer.Id, piwigoServer.Name);
            await AddMissingAlbumsToServerAsync(piwigoClient, piwigoServer, ct);
        }
        else
        {
            _logger.Information("Skipping album creation on server {PiwigoServerId} - {PiwigoServerName}", piwigoServer.Id, piwigoServer.Name);
        }
    }

    public async Task<PendingChange> GetPendingChangesAsync(int piwigoServerId, CancellationToken ct)
    {
        var getOrAdds = await _persistenceContext.PiwigoAlbums.Where(a => a.ServerAlbumId == null && a.ServerId == piwigoServerId).CountAsync(ct);
        const int updates = 0; // not yet supported
        const int deletes = 0; // not yet supported

        return new PendingChange(getOrAdds, updates, deletes);
    }

    private async Task AddMissingAlbumsToServerAsync(IPiwigoClient piwigoClient, ServerEntity piwigoServer, CancellationToken ct)
    {
        var albumsToCreate = await _persistenceContext.PiwigoAlbums.Where(a => a.ServerAlbumId == null && a.ServerId == piwigoServer.Id).OrderBy(a => a.Path).ToListAsync(ct);

        foreach (var albumEntity in albumsToCreate)
        {
            try
            {
                var piwigoParentId = albumEntity.ParentId.HasValue ? (await _persistenceContext.PiwigoAlbums.GetByIdAsync(albumEntity.ParentId.Value, ct)).ServerAlbumId : null;
                _logger.Information("Creating album {AlbumName} ({AlbumPath}) on piwigo server {PiwigoServerName} ({PiwigoServerId})", albumEntity.Name, albumEntity.Path,
                    piwigoServer.Name, piwigoServer.Id);

                albumEntity.ServerAlbumId = await piwigoClient.Album.AddAsync(albumEntity.Name, piwigoParentId, visible: true, position: AlbumPosition.First,
                    status: AlbumStatus.Public, cancellationToken: ct);

                await _persistenceContext.SaveChangesAsync(ct);
            }
            catch (OperationCanceledException)
            {
                _logger.Debug("Cancel adding missing albums to server {PiwigoServerName} ({PiwigoServerId})", piwigoServer.Name, piwigoServer.Id);
                return;
            }
        }
    }

    private async Task UpdatePiwigoAlbumsFromServerAsync(IPiwigoClient piwigoClient, ServerEntity piwigoServer, CancellationToken ct)
    {
        var serverAlbums = await piwigoClient.Album.GetListAsync(null, true, false, ThumbnailSize.Thumb, ct);
        var serverAlbumDictionary = serverAlbums.ToDictionary(a => a.Id, a => a);

        foreach (var serverAlbum in serverAlbums)
        {
            _logger.Information("Updating piwigo server album {ServerAlbumName} with piwigo id {ServerAlbumId}", serverAlbum.Name, serverAlbum.Id);

            var albumEntity = await GetOrAddPiwigoAlbumEntityFromServerAsync(piwigoServer, serverAlbum, serverAlbumDictionary, ct);
            if (serverAlbum.IdUpperCat.HasValue)
            {
                albumEntity.ParentId = (await _persistenceContext.PiwigoAlbums.FindByServerIdAsync(piwigoServer.Id, serverAlbum.IdUpperCat.Value, ct))?.Id;
            }

            await _persistenceContext.SaveChangesAsync(ct);
        }
    }

    private async Task<AlbumEntity> GetOrAddPiwigoAlbumEntityFromServerAsync(ServerEntity piwigoServer, Album serverAlbum, IDictionary<int, Album> serverAlbumDictionary,
        CancellationToken ct)
    {
        // Already synchronized so it is easy to return
        var albumEntity = await _persistenceContext.PiwigoAlbums.FindByServerIdAsync(piwigoServer.Id, serverAlbum.Id, ct);
        if (albumEntity != null)
        {
            _logger.Debug("Found existing album {AlbumEntityName} with local id {AlbumEntityId} and piwigo server id {ServerAlbumId}", albumEntity.Name, albumEntity.Id,
                albumEntity.ServerAlbumId);
            return albumEntity;
        }

        // might exist already as the file system got scanned and created the local entries
        // In this case we save the server id in our local album and link them.
        var path = GeneratePath(serverAlbum, serverAlbumDictionary);
        albumEntity = await _persistenceContext.PiwigoAlbums.FindByServerAndPathAsync(piwigoServer.Id, path, ct);
        if (albumEntity != null)
        {
            albumEntity.ServerAlbumId = serverAlbum.Id;
            _logger.Information("Linking existing album {AlbumEntityName} with local id {AlbumEntityId} to piwigo server id {ServerAlbumId}", albumEntity.Name, albumEntity.Id,
                albumEntity.ServerAlbumId);
            return albumEntity;
        }

        albumEntity = new AlbumEntity
        {
            ServerId = piwigoServer.Id,
            Server = piwigoServer,
            Name = serverAlbum.Name,
            ServerAlbumId = serverAlbum.Id,
            Path = path
        };
        _logger.Information("Adding piwigo album {AlbumEntityName} with local id {AlbumEntityId} and piwigo server id {ServerAlbumId}", albumEntity.Name, albumEntity.Id,
            albumEntity.ServerAlbumId);
        _persistenceContext.PiwigoAlbums.Add(albumEntity);

        return albumEntity;
    }

    private string GeneratePath(Album serverAlbum, IDictionary<int, Album> serverAlbumDictionary)
    {
        if (!serverAlbum.IdUpperCat.HasValue)
        {
            return serverAlbum.Name;
        }

        var parentId = serverAlbum.IdUpperCat.Value;
        var path = serverAlbum.Name;
        while (parentId != 0)
        {
            var currentParent = serverAlbumDictionary[parentId];
            path = Path.Combine(currentParent.Name, path);
            parentId = currentParent.IdUpperCat ?? 0;
        }

        _logger.Debug("Generated path {Path} for album with piwigo id {ServerAlbumId}", path, serverAlbum.Id);

        return path;
    }
}