﻿using System.Collections.Concurrent;
using System.Data;
using System.Threading.Channels;
using Microsoft.EntityFrameworkCore;
using Microsoft.Extensions.DependencyInjection;
using PiwigoDirectorySync.Infrastructure;
using PiwigoDirectorySync.Persistence;
using Polly;
using Serilog;

namespace PiwigoDirectorySync.Services;

internal class FileIndexer : IFileIndexer
{
    private readonly ConcurrentBag<string> _failedFiles = new();
    private readonly ILogger _logger = Log.ForContext<FileIndexer>();

    private readonly IServiceProvider _serviceProvider;
    private int _totalFilesScanned;

    public FileIndexer(IServiceProvider serviceProvider)
    {
        _serviceProvider = serviceProvider;
    }

    public int TotalFilesScanned => _totalFilesScanned;

    public IReadOnlyCollection<string> FailedFiles => _failedFiles;

    public async Task StartProcessingAsync(Channel<string> fileQueue, int piwigoServerId, CancellationToken ct)
    {
        var parallelOptions = new ParallelOptions
        {
            MaxDegreeOfParallelism = Math.Min(Environment.ProcessorCount, 8),
            CancellationToken = ct
        };
        await Parallel.ForEachAsync(fileQueue.Reader.ReadAllAsync(ct), parallelOptions,
            async (fullFilePath, token) => { await ProcessFileAsync(fullFilePath, piwigoServerId, token); });
    }

    private async Task ProcessFileAsync(string fullFilePath, int piwigoServerId, CancellationToken ct)
    {
        try
        {
            if (ct.IsCancellationRequested)
            {
                _logger.Warning("Indexing cancelled");
                return;
            }

            _logger.Debug("Indexing file {FullFilePath}", fullFilePath);
            var fileInfo = new FileInfo(fullFilePath);

            if (!fileInfo.Exists)
            {
                _logger.Warning("File {FullFilePath} not found", fullFilePath);
                _failedFiles.Add(fullFilePath);
                return;
            }

            await using var scope = _serviceProvider.CreateAsyncScope();
            await using var db = scope.ServiceProvider.GetRequiredService<PersistenceContext>();
            var piwigoServer = await db.PiwigoServers.GetByIdAsync(piwigoServerId, ct);

            var relativePath = Path.GetRelativePath(piwigoServer.RootDirectory, fullFilePath);

            var album = await GetOrAddAlbumAsync(piwigoServerId, fileInfo.Directory!, ct);

            var image = await GetOrAddImageAsync(db, album, relativePath, ct);

            if (image.LastChange != fileInfo.LastWriteTimeUtc)
            {
                _logger.Information("Calculating md5 sum for file {FullFilePath}", fullFilePath);
                image.UploadRequired = true;
                image.Md5Sum = await FilesystemHelpers.CalculateMd5SumAsync(fullFilePath, ct);
            }

            image.DeleteRequired = false;
            image.LastChange = fileInfo.LastWriteTimeUtc;

            await db.SaveChangesAsync(ct);

            Interlocked.Add(ref _totalFilesScanned, 1);
        }
        catch (Exception ex)
        {
            _failedFiles.Add(fullFilePath);
            _logger.Error(ex, "could not index file {FullFilePath}", fullFilePath);
        }
    }

    private async Task<ImageEntity> GetOrAddImageAsync(PersistenceContext db, int albumId, string relativePath, CancellationToken ct)
    {
        var imageEntity = await db.PiwigoImages.Where(i => i.AlbumId == albumId && i.FilePath == relativePath).FirstOrDefaultAsync(ct);
        if (imageEntity is not null)
        {
            _logger.Debug("Found file {RelativePath}", relativePath);
            return imageEntity;
        }

        _logger.Information("Adding new file {RelativePath}", relativePath);
        imageEntity = new ImageEntity
        {
            AlbumId = albumId,
            FilePath = relativePath,
            UploadRequired = true,
            DeleteRequired = false
        };
        db.PiwigoImages.Add(imageEntity);

        return imageEntity;
    }

    private async Task<int> GetOrAddAlbumAsync(int serverId, DirectoryInfo directory, CancellationToken ct)
    {
        return await Policy.Handle<DbUpdateException>()
            .RetryAsync(3, (ex, _) => { _logger.Warning(ex, "Could not create album for directory {AlbumDirectory}, retrying", directory); })
            .ExecuteAsync(async () =>
            {
                await using var scope = _serviceProvider.CreateAsyncScope();
                await using var db = scope.ServiceProvider.GetRequiredService<PersistenceContext>();
                await using var tx = await db.Database.BeginTransactionAsync(IsolationLevel.Serializable, ct);
                var piwigoServer = await db.PiwigoServers.GetByIdAsync(serverId, ct);

                var albumId = await GetOrAddAlbumAsync(db, serverId, piwigoServer.RootDirectory, directory, ct);

                await tx.CommitAsync(ct);
                return albumId;
            });
    }

    private async Task<int> GetOrAddAlbumAsync(PersistenceContext db, int serverId, string rootDirectory, DirectoryInfo directory, CancellationToken ct)
    {
        var albumPath = Path.GetRelativePath(rootDirectory, directory.FullName);
        var album = await db.PiwigoAlbums.FindByServerAndPathAsync(serverId, albumPath, ct);
        if (album != null)
        {
            _logger.Debug("Found album with id {AlbumId} for directory {AlbumDirectory}", album.Id, directory);
            return album.Id;
        }

        _logger.Information("creating album for directory {AlbumDirectory}", directory);
        int? parentAlbumId = null;
        if (!string.Equals(new DirectoryInfo(rootDirectory).FullName, directory.Parent!.FullName))
        {
            parentAlbumId = await GetOrAddAlbumAsync(db, serverId, rootDirectory, directory.Parent!, ct);
        }

        album = new AlbumEntity
        {
            ServerId = serverId,
            Name = directory.Name,
            Path = albumPath,
            ParentId = parentAlbumId
        };
        db.PiwigoAlbums.Add(album);

        await db.SaveChangesAsync(ct);

        return album.Id;
    }
}