using System.Threading.Channels;
using PiwigoDirectorySync.Persistence;
using Serilog;

namespace PiwigoDirectorySync.Services;

internal class FileSystemScanner : IFileSystemScanner
{
    private readonly ILogger _logger = Log.ForContext<FileSystemScanner>();
    private readonly PersistenceContext _persistenceContext;

    public FileSystemScanner(PersistenceContext persistenceContext)
    {
        _persistenceContext = persistenceContext;
    }

    public async Task ScanAsync(Channel<string> fileQueue, int piwigoServerId, CancellationToken ct)
    {
        try
        {
            var piwigoServer = await _persistenceContext.PiwigoServers.GetByIdAsync(piwigoServerId, ct);
            _logger.Information("Scanning files for piwigo server {PiwigoServerName} in directory {PiwigoServerRootDirectory}", piwigoServer.Name, piwigoServer.RootDirectory);

            await ScanRootDirectory(fileQueue, new DirectoryInfo(piwigoServer.RootDirectory), ct);
        }
        finally
        {
            fileQueue.Writer.Complete();
        }
    }

    private async ValueTask ScanRootDirectory(Channel<string> fileQueue, DirectoryInfo directory, CancellationToken ct)
    {
        _logger.Information("Scanning root directory {DirectoryFullName} for sidecars to delete", directory.FullName);
        var parallelOptions = new ParallelOptions
        {
            CancellationToken = ct
        };
        await Parallel.ForEachAsync(GetDirectories(directory), parallelOptions, (d, c) => FindAndEnqueueFilesToAdd(fileQueue, d, c));
    }

    private async ValueTask FindAndEnqueueFilesToAdd(Channel<string> fileQueue, DirectoryInfo directory, CancellationToken ct)
    {
        try
        {
            _logger.Information("Scanning directory {DirectoryFullName} for images", directory.FullName);

            var imageFiles = AppSettings.SupportedExtensions.SelectMany(ext => directory.GetFiles($"*.{ext}", SearchOption.TopDirectoryOnly));
            foreach (var imageFile in imageFiles)
            {
                _logger.Debug("Found image {ImageFileFullName}, enqueue index", imageFile.FullName);
                await fileQueue.Writer.WriteAsync(imageFile.FullName, ct);
            }
        }
        catch (Exception ex)
        {
            _logger.Error(ex, "could not scan directory {DirectoryFullName}", directory.FullName);
        }
    }

    private static IEnumerable<DirectoryInfo> GetDirectories(DirectoryInfo directoryInfo)
    {
        yield return directoryInfo;
        foreach (var directory in directoryInfo.EnumerateDirectories().SelectMany(GetDirectories).OrderBy(d => d.FullName))
        {
            yield return directory;
        }
    }
}