﻿using Flurl.Http;
using Microsoft.EntityFrameworkCore;
using Piwigo.Client;
using Piwigo.Client.Albums;
using Piwigo.Client.Images;
using PiwigoDirectorySync.Infrastructure;
using PiwigoDirectorySync.Persistence;
using Serilog;

namespace PiwigoDirectorySync.Services;

internal class ImageSynchronizer : IImageSynchronizer
{
    private readonly ILogger _logger = Log.ForContext<ImageSynchronizer>();
    private readonly PersistenceContext _persistenceContext;
    private readonly IPiwigoClientFactory _piwigoClientFactory;

    public ImageSynchronizer(IPiwigoClientFactory piwigoClientFactory, PersistenceContext persistenceContext)
    {
        _piwigoClientFactory = piwigoClientFactory;
        _persistenceContext = persistenceContext;
    }

    public async Task<PendingChange> GetPendingChangesAsync(int piwigoServerId, CancellationToken ct)
    {
        var getOrAdds = await _persistenceContext.PiwigoImages.Where(i => i.Album.ServerId == piwigoServerId && i.ServerImageId == null).CountAsync(ct);
        var updates = await _persistenceContext.PiwigoImages.Where(i => i.Album.ServerId == piwigoServerId && i.ServerImageId != null && i.UploadRequired).CountAsync(ct);
        var deletes = await _persistenceContext.PiwigoImages.Where(i => i.Album.ServerId == piwigoServerId && i.ServerImageId != null && i.DeleteRequired).CountAsync(ct);

        return new PendingChange(getOrAdds, updates, deletes);
    }

    public async Task DownloadImagesAsync(int piwigoServerId, CancellationToken ct)
    {
        var piwigoServer = await _persistenceContext.PiwigoServers.FindAsync(new object?[] { piwigoServerId }, ct);
        if (piwigoServer is null)
        {
            _logger.Error("Could not sync images with piwigo server {PiwigoServerId}", piwigoServerId);
            return;
        }

        _logger.Information("Downloading missing images of piwigo server {PiwigoServerName} using base path {PiwigoServerRootDirectory}", piwigoServer.Name,
            piwigoServer.RootDirectory);

        var piwigoClient = await _piwigoClientFactory.GetPiwigoClientAsync(piwigoServer, ct);

        var albumIdsToDownload = await _persistenceContext.PiwigoAlbums.Where(a => a.ServerAlbumId.HasValue).Select(a => a.ServerAlbumId!.Value).Distinct().ToListAsync(ct);

        foreach (var albumId in albumIdsToDownload)
        {
            var albumInfos = await piwigoClient.Album.GetListAsync(albumId, false, false, ThumbnailSize.Thumb, ct);
            if (!albumInfos.Any())
            {
                Log.Warning("Album with server {AlbumId} not found on piwigo server", albumId);
                continue;
            }

            var albumInfo = albumInfos.First();

            _logger.Information("Starting downloads for album {AlbumInfoName}", albumInfo.Name);

            await DownloadImagesForAlbumAsync(piwigoClient, piwigoServer, albumId, albumInfo, ct);
        }
    }

    public async Task SynchronizeImagesAsync(int piwigoServerId, bool readOnly, CancellationToken ct)
    {
        var piwigoServer = await _persistenceContext.PiwigoServers.FindAsync(new object?[] { piwigoServerId }, ct);
        if (piwigoServer is null)
        {
            _logger.Error("Could not sync images with piwigo server {PiwigoServerId}", piwigoServerId);
            return;
        }

        _logger.Information("Synchronizing images of piwigo server {PiwigoServerName} using base path {PiwigoServerRootDirectory}", piwigoServer.Name, piwigoServer.RootDirectory);

        var piwigoClient = await _piwigoClientFactory.GetPiwigoClientAsync(piwigoServer, ct);

        await GetImageIdsFromServerAsync(piwigoClient, piwigoServer, ct);
        if (!readOnly)
        {
            await UploadNewImagesToServerAsync(piwigoClient, piwigoServer, ct);
            await UploadChangedImagesToServerAsync(piwigoClient, piwigoServer, ct);
        }
        else
        {
            _logger.Information("Skipping image upload of server {PiwigoServerName} as readOnly is specified", piwigoServer.Name);
        }
    }

    private async Task DownloadImagesForAlbumAsync(IPiwigoClient piwigoClient, ServerEntity piwigoServer, int albumId, Album albumInfo, CancellationToken ct)
    {
        if (albumInfo.NbImages is null or <= 0)
        {
            _logger.Information("No images to download for empty album {AlbumId} / {AlbumInfoName}", albumId, albumInfo.Name);
            return;
        }

        const int pageSize = 100;
        var numberOfPages = albumInfo.NbImages.Value / pageSize + 1;
        var currentPage = 0;

        while (currentPage < numberOfPages)
        {
            var imagePagingInfo = new ImagePagingInfo(currentPage, pageSize, albumInfo.NbImages.Value);
            var images = await piwigoClient.Image.GetImagesAsync(albumId, false, imagePagingInfo, ImageFilter.Empty, ImageOrder.Name, ct);

            foreach (var image in images.Images)
            {
                try
                {
                    await DownloadImageAsync(piwigoServer, albumInfo, image, ct);
                }
                catch (OperationCanceledException)
                {
                    Log.Debug("Cancel of download requested");
                    return;
                }
                catch (Exception ex)
                {
                    Log.Warning(ex, "Download of server image {ImageId} / {ImageName} / {ImageElementUrl}", image.Id, image.Name, image.ElementUrl);
                }
            }

            currentPage++;
        }
    }

    private async Task DownloadImageAsync(ServerEntity piwigoServer, Album albumInfo, Image image, CancellationToken ct)
    {
        var localAlbum = await _persistenceContext.PiwigoAlbums.FindByServerIdAsync(piwigoServer.Id, albumInfo.Id, ct);
        if (localAlbum is null)
        {
            _logger.Warning("Could not add image {ImageId} / {ImageFile}: album with server id {AlbumInfoId} / {AlbumInfoName} not found", image.Id, image.File, albumInfo.Id,
                albumInfo.Name);
            return;
        }

        var localImage = await GetOrAddImageFromServerAsync(localAlbum, image, ct);

        var fileInfo = new FileInfo(Path.Combine(piwigoServer.RootDirectory, localImage.FilePath));
        if (fileInfo.Exists)
        {
            _logger.Debug("Tried to download image {ImageFile} but it already exists", fileInfo.FullName);
            return;
        }

        _logger.Information("Downloading image {ImageFile}", fileInfo.FullName);
        await image.ElementUrl.DownloadFileAsync(fileInfo.Directory!.FullName, fileInfo.Name, cancellationToken: ct);
        localImage.Md5Sum = await FilesystemHelpers.CalculateMd5SumAsync(fileInfo.FullName, ct);
        await _persistenceContext.SaveChangesAsync(ct);
    }

    private async Task<ImageEntity> GetOrAddImageFromServerAsync(AlbumEntity album, Image image, CancellationToken ct)
    {
        var imageFullPath = Path.Combine(album.Path, image.File!);
        var imageEntity = await _persistenceContext.PiwigoImages.Where(i => i.AlbumId == album.Id && i.ServerImageId == image.Id).FirstOrDefaultAsync(ct);
        if (imageEntity is not null)
        {
            return imageEntity;
        }

        imageEntity = await _persistenceContext.PiwigoImages.Where(i => i.AlbumId == album.Id && i.FilePath == imageFullPath).FirstOrDefaultAsync(ct);
        if (imageEntity is not null)
        {
            imageEntity.ServerImageId = image.Id;
            return imageEntity;
        }

        imageEntity = new ImageEntity
        {
            AlbumId = album.Id,
            Album = album,
            ServerImageId = image.Id,
            FilePath = imageFullPath,
            UploadRequired = false,
            DeleteRequired = false
        };
        _persistenceContext.PiwigoImages.Add(imageEntity);
        return imageEntity;
    }

    private async Task UploadChangedImagesToServerAsync(IPiwigoClient piwigoClient, ServerEntity piwigoServer, CancellationToken ct)
    {
        var imagesToUpload = await _persistenceContext.PiwigoImages.Include(i => i.Album)
            .Where(i => i.ServerImageId != null && i.Album.ServerId == piwigoServer.Id && i.UploadRequired && i.ServerImageId != null)
            .OrderBy(i => i.FilePath)
            .ToListAsync(ct);

        _logger.Information("Updating {Count} images", imagesToUpload.Count);

        foreach (var imageEntity in imagesToUpload)
        {
            try
            {
                var fileInfo = new FileInfo(Path.Combine(piwigoServer.RootDirectory, imageEntity.FilePath));

                var fileSizeInKb = fileInfo.Length / 1024f;
                _logger.Information("Updating image {ImageFilePath} ({ImageId}) to piwigo server ({Size} KiB)", imageEntity.FilePath, imageEntity.ServerImageId, fileSizeInKb);

                var imageUpload = GetImageUpload(imageEntity, fileInfo.CreationTime);
                var imageUploaded = await piwigoClient.UploadImageAsync(fileInfo, imageUpload, ct);

                imageEntity.ServerImageId = imageUploaded.ImageId;
                imageEntity.UploadRequired = false;

                _logger.Information("Updated image {ImageEntityPath} ({ImageEntityId}) on piwigo server with id {ImageEntityServerImageId}", imageEntity.FilePath, imageEntity.Id,
                    imageEntity.ServerImageId);

                await _persistenceContext.SaveChangesAsync(ct);
            }
            catch (OperationCanceledException)
            {
                Log.Debug("Cancel of image upload requested");
                return;
            }
            catch (PiwigoException ex)
            {
                _logger.Error(ex, "Could not update image {ImageEntityPath} ({ImageEntityId}) on piwigo server with id {ImageEntityServerImageId}", imageEntity.FilePath,
                    imageEntity.Id, imageEntity.ServerImageId);
            }
        }
    }

    private async Task UploadNewImagesToServerAsync(IPiwigoClient piwigoClient, ServerEntity piwigoServer, CancellationToken ct)
    {
        var imagesToUpload = await _persistenceContext.PiwigoImages.Include(i => i.Album)
            .Where(i => i.ServerImageId == null && i.Album.ServerId == piwigoServer.Id && i.Album.ServerAlbumId != null)
            .OrderBy(i => i.FilePath)
            .ToListAsync(ct);

        _logger.Information("Uploading {Count} images", imagesToUpload.Count);

        foreach (var imageEntity in imagesToUpload)
        {
            try
            {
                var fileInfo = new FileInfo(Path.Combine(piwigoServer.RootDirectory, imageEntity.FilePath));

                var fileSizeInKb = fileInfo.Length / 1024f;
                _logger.Information("Uploading image {ImageFilePath} ({ImageId}) to piwigo server ({Size} KiB)", imageEntity.FilePath, imageEntity.Id, fileSizeInKb);

                var imageUpload = GetImageUpload(imageEntity, fileInfo.CreationTime);
                var imageUploaded = await piwigoClient.UploadImageAsync(fileInfo, imageUpload, ct);

                imageEntity.ServerImageId = imageUploaded.ImageId;
                imageEntity.UploadRequired = false;

                _logger.Information("Uploaded image {ImageFilePath} ({ImageId}) to piwigo server with id {ImageServerImageId}", imageEntity.FilePath, imageEntity.Id,
                    imageEntity.ServerImageId);

                await _persistenceContext.SaveChangesAsync(ct);
            }
            catch (OperationCanceledException)
            {
                Log.Debug("Cancel of image upload requested");
                return;
            }
            catch (PiwigoException ex)
            {
                _logger.Error(ex, "Could not upload image {ImageEntityPath} ({ImageEntityId}) on piwigo server with id {ImageEntityServerImageId}", imageEntity.FilePath,
                    imageEntity.Id, imageEntity.ServerImageId);
            }
        }
    }

    private static ImageUpload GetImageUpload(ImageEntity imageEntity, DateTime createdAt)
    {
        var albums = new List<(int AlbumId, int? Rank)> { (AlbumId: imageEntity.Album.ServerAlbumId!.Value, Rank: null) };
        return new ImageUpload(imageEntity.Md5Sum!)
        {
            Name = imageEntity.Name,
            FileName = imageEntity.FileName,
            OriginalSum = imageEntity.Md5Sum!,
            Albums = albums,
            CreatedAt = createdAt
        };
    }

    private async Task GetImageIdsFromServerAsync(IPiwigoClient piwigoClient, ServerEntity piwigoServer, CancellationToken ct)
    {
        var imagesToSearch = await _persistenceContext.PiwigoImages.Include(i => i.Album)
            .Where(i => i.ServerImageId == null && i.Album.ServerId == piwigoServer.Id)
            .OrderBy(i => i.FilePath)
            .ToListAsync(ct);

        _logger.Information("Checking {Count} images if they exist", imagesToSearch.Count);

        var md5SumsToCheck = imagesToSearch.Where(i => i.Md5Sum != null).DistinctBy(i => i.Md5Sum).ToDictionary(i => i.Md5Sum!, i => i, StringComparer.OrdinalIgnoreCase);

        var processedImages = 0;
        while (processedImages < md5SumsToCheck.Keys.Count)
        {
            var chunkKeys = md5SumsToCheck.Keys.Skip(processedImages).Take(100).ToArray();
            var existingImages = await piwigoClient.Image.ExistsByMd5SumsAsync(chunkKeys, ct);

            foreach (var existingImage in existingImages.Where(i => i.Value.HasValue))
            {
                var imageEntity = md5SumsToCheck[existingImage.Key];
                imageEntity.ServerImageId = existingImage.Value;
                imageEntity.UploadRequired = false;

                _logger.Information("Found image {ImageEntityName} ({ImageEntityId}) on piwigo server with id {ImageEntityServerImageId}", imageEntity.Name, imageEntity.Id,
                    imageEntity.ServerImageId);
            }

            await _persistenceContext.SaveChangesAsync(ct);

            processedImages += chunkKeys.Length;
        }
    }
}